@@ -1,3 +1,4 @@
+from .version import *
 from .config import *
 from .base import *
 from .data_distributions import *
@@ -7,3 +8,6 @@ from .contact_networks import *
 from .contact_networks_industries import *
 from .plot_tools import *
 from .api import *
+from .long_term_care_facilities import *
+from .school_modules import *
+from .read_write import *
